from abc import ABC, abstractclassmethod
from inspect import iscoroutinefunction
from typing import Awaitable, Callable

from promplate.llm.base import *
from promplate.prompt import ChatTemplate, Template
from promplate.prompt.template import Context

Process = Callable[[Context], Context]

AsyncProcess = Callable[[Context], Awaitable[Context]]


class AbstractChain(ABC):
    @abstractclassmethod
    def run(self, context: Context, complete: Complete):
        pass

    @abstractclassmethod
    async def arun(self, context: Context, complete: Complete | AsyncComplete):
        pass


class Node(AbstractChain):
    def __init__(
        self,
        template: Template,
        pre_processes: list[Process | AsyncProcess] | None = None,
        post_processes: list[Process | AsyncProcess] | None = None,
        **config,
    ):
        self.template = template
        self.pre_processes = pre_processes or []
        self.post_processes = post_processes or []
        self.run_config = config

    def run(self, context, complete):
        for process in self.pre_processes:
            context = process(context)

        prompt = self.template.render(context)

        assert isinstance(self.template, ChatTemplate) ^ isinstance(prompt, str)

        context["__result__"] = complete(prompt, **self.run_config)

        for process in self.post_processes:
            context = process(context)

        return context

    async def arun(self, context, complete):
        for process in self.pre_processes:
            if iscoroutinefunction(process):
                context = await process(context)
            else:
                context = process(context)

        prompt = await self.template.arender(context)

        assert isinstance(self.template, ChatTemplate) ^ isinstance(prompt, str)

        if iscoroutinefunction(complete):
            context["__result__"] = await complete(prompt, **self.run_config)
        else:
            context["__result__"] = complete(prompt, **self.run_config)

        for process in self.post_processes:
            if iscoroutinefunction(process):
                context = await process(context)
            else:
                context = process(context)

        return context

    def next(self, chain: AbstractChain):
        if isinstance(chain, Node):
            return Chain(self, chain)
        else:
            return Chain(self, *chain)

    def __add__(self, chain: AbstractChain):
        return self.next(chain)


class Chain(AbstractChain):
    def __init__(self, *nodes: AbstractChain):
        self.nodes = list(nodes)

    def next(self, chain: AbstractChain):
        if isinstance(chain, Node):
            return Chain(*self, chain)
        else:
            return Chain(*self, *chain)

    def __add__(self, chain):
        return self.next(chain)

    def __iter__(self):
        return self.nodes

    def run(self, context, complete):
        for node in self.nodes:
            context = node.run(context, complete)

        return context

    async def arun(self, context, complete):
        for node in self.nodes:
            context = await node.arun(context, complete)

        return context



from openai import ChatCompletion, Completion

from promplate.prompt.chat import Message


def text_complete(**default_config):
    def complete_text(text: str, **config) -> str:
        config = default_config | config
        res = Completion.create(prompt=text, **config)
        return res["choices"][0]["text"]

    return complete_text


def async_text_complete(**default_config):
    async def async_complete_text(text: str, **config) -> str:
        config = default_config | config
        res = await Completion.acreate(prompt=text, **config)
        return res["choices"][0]["text"]

    return async_complete_text


def chat_complete(**default_config):
    def complete_chat(messages: list[Message], **config) -> str:
        config = default_config | config
        res = ChatCompletion.create(messages=messages, **config)
        return res["choices"][0]["message"]["content"]

    return complete_chat


def async_chat_complete(**default_config):
    async def async_complete_chat(messages: list[Message], **config) -> str:
        config = default_config | config
        res = await ChatCompletion.acreate(messages=messages, **config)
        return res["choices"][0]["message"]["content"]

    return async_complete_chat



from typing import Awaitable, Callable

from promplate.prompt.chat import Message

CompleteText = Callable[[str], str]
CompleteChat = Callable[[list[Message]], str]
Complete = CompleteText | CompleteChat

AsyncCompleteText = Callable[[str], Awaitable[str]]
AsyncCompleteChat = Callable[[list[Message]], Awaitable[str]]
AsyncComplete = AsyncCompleteText | AsyncCompleteChat
